@@ -201,11 +201,7 @@ listen_tcp(PortIP, Module, SockOpts, Port, IPS) ->
 			catch
 			    _:_ -> []
 			end,
-	    DeliverAs = case Module of
-			    ejabberd_xmlrpc -> list;
-			    _ -> binary
-			end,
-	    Res = gen_tcp:listen(Port, [DeliverAs,
+	    Res = gen_tcp:listen(Port, [binary,
 					{packet, 0},
 					{active, false},
 					{reuseaddr, true},
@@ -595,7 +591,7 @@ transform_option({{Port, IP, Transport}, Mod, Opts}) ->
                       try
                           Mod:transform_listen_option(Opt, Acc)
                       catch error:undef ->
-                              Acc
+                              [Opt|Acc]
                       end
               end, [], Opts1),
     TransportOpt = if Transport == tcp -> [];
