@@ -374,8 +374,8 @@ wait_for_feature_request({xmlstreamelement, El},
 				    #xmlel{name = <<"success">>,
 					   attrs = [{<<"xmlns">>, ?NS_SASL}],
 					   children = []}),
-		       ?DEBUG("(~w) Accepted s2s authentication for ~s",
-			      [StateData#state.socket, AuthDomain]),
+		       ?INFO_MSG("Accepted s2s EXTERNAL authentication for ~s (TLS=~p)",
+				 [AuthDomain, StateData#state.tls_enabled]),
 		       change_shaper(StateData, <<"">>,
 				     jlib:make_jid(<<"">>, AuthDomain, <<"">>)),
 		       {next_state, wait_for_stream,
@@ -515,6 +515,8 @@ stream_established({valid, From, To}, StateData) ->
 			    [{<<"from">>, To}, {<<"to">>, From},
 			     {<<"type">>, <<"valid">>}],
 			children = []}),
+    ?INFO_MSG("Accepted s2s dialback authentication for ~s (TLS=~p)",
+	      [From, StateData#state.tls_enabled]),
     LFrom = jlib:nameprep(From),
     LTo = jlib:nameprep(To),
     NSD = StateData#state{connections =
